/**
 ** MPC_DEMTreatmentExecTaskCodels.cc
 **
 ** Codels called by execution task MPC_DEMTreatmentExecTask
 **
 ** Author: Vivien Delsart
 ** Date: 
 **
 **/

#include <portLib.h>

#include "server/MPC_DEMTreatmentHeader.h"

#include "DEMTTT_Posters_ttt.h"

#include <MPC/common.hpp>
#include <MPC/environment.hpp>
using namespace mpc;

#include <iostream>
#include <cstring>
#include <list>
using namespace std;

#define DEBUG 0

DEM* dem_;                                /**< Data Elevation Map on which the robot is navigating */
VarSlopeMap* slopeMap_;                   /**< Slope map corresponding to the DEM */
DistObsMap* distMap_;                     /**< Distance to obstacles map corresponding to the DEM */

POSTER_ID demPosterId;  

std::list<DEM*> logDEM;                   /**< List containing the DEM log */

Timer timer_monitor;                 // timer used to monitor the real execution time of the module


/*------------------------------------------------------------------------
 *
 * MPC_DT_ExecTaskInit  --  Initialization codel (fIDS, ...)
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
MPC_DT_ExecTaskInit(int *report)
{
  demPosterId = NULL;

  dem_ = NULL;
  slopeMap_ = NULL;
  distMap_ = NULL;

  SDI_F->timer_started = 0;

  SDI_F->DT_slopeMapDist = 0.3;
  SDI_F->DT_thresSlope = 0.25 * M_PI;
  SDI_F->DT_thresVarSlope = 5.0 * M_PI;
  SDI_F->DT_safeDist = 0.1;
  SDI_F->DT_unkwnIsObs = 0;

  SDI_F->logDEM = 0;
  SDI_F->display_time = 0;

  logDEM.clear();

  return OK;
}


/*------------------------------------------------------------------------
 *
 * MPC_DT_ExecTaskEnd  --  Termination codel
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
MPC_DT_ExecTaskEnd(void)
{
  return OK;
}


/*------------------------------------------------------------------------
 *
 * MPC_DT_ExecTaskPerm  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: 
 * 
 * Reports:      OK
 * 
 * Returns:    OK or ERROR
 */

STATUS
MPC_DT_ExecTaskPerm(int *report)
{
  return OK;
}


/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_DEMTreatment_DEM_POSTER_NOT_FOUND
 */

/* MPC_DT_InitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_DT_InitMain(DT_INIT_I *Init_I, int *report)
{
  // check for the Data Elevation Map poster (DTM)
  if (posterFind(Init_I->demPoster, &demPosterId) == ERROR) 
    {
      cerr << "MPC_DEMTreatment: cannot find dem poster" << endl;
      *report = S_MPC_DEMTreatment_DEM_NOT_FOUND;
      return ETHER;
    }
#if DEBUG
  else cerr << "MPC_DEMTreatment: dem poster found" << endl;
#endif

  SDI_F->distObsMap.dimX = 0;
  SDI_F->distObsMap.dimY = 0;
  SDI_F->distObsMap.thresSlope = 0;
  SDI_F->distObsMap.thresVarSlope = 0;
  SDI_F->distObsMap.safeDist = 0;

  *report = OK;
  return ETHER;
}


/*------------------------------------------------------------------------
 * Compute_DistMap
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_DEMTreatment_DEM_NOT_FOUND
 */

/* Compute_DistMapMain  -  codel EXEC of Compute_DistMap
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
Compute_DistMapMain(BUILD_DISTMAP_I *BuildDistMap_I, int *report)
{
  SDI_F->DT_slopeMapDist = BuildDistMap_I->slopeMapDist;
  SDI_F->DT_thresSlope = BuildDistMap_I->thresSlope;
  SDI_F->DT_thresVarSlope = BuildDistMap_I->thresVarSlope;
  SDI_F->DT_safeDist = BuildDistMap_I->safeDist;
  
  SDI_F->DT_unkwnIsObs = BuildDistMap_I->unkwnIsObs;

  // check if the DEM poster has been found
  if(demPosterId == NULL)
    {
      cerr << "MPC_DEMTreatment: Cannot start the distmap building process : DEM poster has not been found." << endl;      
      return ETHER;
    }

  // update the robot local DEM
  if(dem_ != NULL) delete dem_;
  dem_ = updateDEM(demPosterId); 
  if(dem_ == NULL)
    {
      cerr << "MPC_DEMTreatment : DEM UPDATE FAILED" << endl;
      return ETHER;
    }

  // compute a slope map from the input DEM
  if(slopeMap_ != NULL) delete slopeMap_;
  slopeMap_ = new VarSlopeMap(dem_, SDI_F->DT_slopeMapDist);
  
  // init the distmap computation options
  if(SDI_F->DT_unkwnIsObs == 1) DistObsMap::unkwnIsObs_ = true;
  else DistObsMap::unkwnIsObs_ = false;
  
  // compute a distance to obstacles map from the slopemap
  if(distMap_ != NULL) delete distMap_;
  distMap_ = new DistObsMap(slopeMap_, 
			    SDI_F->DT_thresSlope, 
			    SDI_F->DT_thresVarSlope, 
			    SDI_F->DT_safeDist);
  
  // export the distmap in a poster
  int res = fillDistMapPoster(distMap_);
  
  if(res != 0) 
    {
      cerr << "DEM_Treatment : Unable to fill the Dist Map poster. " << endl;
      return ETHER;
    }

  // DEM log
  if(SDI_F->logDEM == 1)
    {
      // create a copy of the current DEM
      DEM* copyDEM = dem_->copy();
      
      // append it to the DEM log
      logDEM.push_back(copyDEM);
    }

  if(SDI_F->display_time && SDI_F->timer_started == 1)
    {
      cout << "MPC_DEMTreatment::MPC_DT_ExecTaskPerm -> Elapsed time : " << timer_monitor.elapsedTime() << "   period : " << timer_monitor.avgPeriod() << "   frequency : " << timer_monitor.avgFrequency() << endl;
    }

  return ETHER;
}


/*------------------------------------------------------------------------
 * Merge_DEMLog
 *
 * Description: 
 *
 * Reports:      OK
 */

/* Merge_DEMLogMain  -  codel EXEC of Merge_DEMLog
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
Merge_DEMLogMain(MERGE_DEMLOG_I *Merge_DEMLog_I, int *report)
{
  // check if the DEM log is not empty
  if(logDEM.size() == 0)
    {
      cerr << "Cannot save a DEM. The DEM log is empty." << endl;
      return ETHER;
    }

  // check for the (topmost leftmost), the (downmost rightmost) logged DEMs and for their dimensions
  float max_up = 0;
  float max_left = 0;
  float max_down = 0;
  float max_right = 0;
  float zOrigin = 0;
 
  float xScale = 0;
  float yScale = 0;
  float zScale = 0;
  int nbCols = 0;
  int nbLines = 0;

  for(std::list<DEM*>::iterator it = logDEM.begin(); it != logDEM.end(); it++)
    {
      // get back the current DEM 
      float xOrigin = (*it)->xOrigin();
      float yOrigin = (*it)->yOrigin();

      if(it == logDEM.begin())
	{
	  max_up = yOrigin;
	  max_left = xOrigin;
	  max_down = yOrigin;
	  max_right = xOrigin;
	  zOrigin = (*it)->zOrigin();

	  xScale = (*it)->xScale();
	  yScale = (*it)->yScale();
	  zScale = (*it)->zScale();
	  nbCols = (*it)->nbCols();
	  nbLines = (*it)->nbLines();
	}
      else
	{
	  if(xOrigin < max_left) max_left = xOrigin;
	  else if(xOrigin > max_right) max_right = xOrigin;

	  if(yOrigin < max_up) max_up = yOrigin;
	  else if(yOrigin > max_down) max_down = yOrigin;
	}
    }

//   cout << "max_up : " << max_up << "  max_down : " << max_down << "  max_left : " << max_left << "  max_right : " << max_right << endl;

  // compute the dimension of the new grid
  float dx = max_right - max_left;
  float dy = max_down - max_up;

//   cout << "dx : " << dx << "  dy : " << dy << endl;

  int newNbCols = (dx / xScale) + nbCols + 1;
  int newNbLines = (dy / yScale) + nbLines + 1;
  if(newNbCols >= newNbLines) newNbLines = newNbCols;
  else newNbCols = newNbLines;

  // create a new empty DEM with the dimensions found
  DEM* mergeDEM = new DEM(newNbCols, newNbLines, max_left, max_up, zOrigin, xScale, yScale, zScale);

  cout << "newNbCols : " << newNbCols << "  newNbLines : " << newNbLines << "  max_left : " << max_left << "  max_up : " << max_up << "  zOrigin : " << zOrigin << "  xScale : " << xScale << "  yScale : " << yScale << "  zScale : " << zScale << endl;

  cout << "mergeDEM->nbCols() : " << mergeDEM->nbCols() << endl;
  cout << "mergeDEM->nbLines() : " << mergeDEM->nbLines() << endl;
  cout << "mergeDEM->xOrigin() : " << mergeDEM->xOrigin() << endl;
  cout << "mergeDEM->yOrigin() : " << mergeDEM->yOrigin() << endl;
  cout << "mergeDEM->xScale() : " << mergeDEM->xScale() << endl;
  cout << "mergeDEM->yScale() : " << mergeDEM->yScale() << endl;
  
  int i=0;
  for(std::list<DEM*>::iterator it = logDEM.begin(); it != logDEM.end(); it++)
    {
      int resMerge = mergeDEM->mergeDEM(**it);
      if(resMerge != 0) cerr << "MPC_DEMTreatment : The DEM merging process failed :-(" << endl;
    }

  // save the merged DEM into an output file
  int res_save = mergeDEM->save(Merge_DEMLog_I->filename);
  if(res_save != 0) cerr << "MPC_DEMTreatment : Cannot save the merged DEM (log)." << endl;

  int res_save2 = mergeDEM->saveIntoPPM("mergeDEM.ppm");
  if(res_save != 0) cerr << "MPC_DEMTreatment : Cannot save the merged DEM (ppm)." << endl;

  // delete the merged DEM
  if(mergeDEM != NULL) delete mergeDEM;
  
  return ETHER;
}


/*------------------------------------------------------------------------
 * Clear_DEMLog
 *
 * Description: 
 *
 * Reports:      OK
 */

/* Clear_DEMLogMain  -  codel EXEC of Clear_DEMLog
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
Clear_DEMLogMain(int *report)
{
  // delete every DEM logged
  for(std::list<DEM*>::iterator it = logDEM.begin(); it != logDEM.end(); it++)
    {
      if(*it != NULL) delete *it;
    }
  
  // clear the log list
  logDEM.clear();

  return ETHER;
}


/*------------------------------------------------------------------------
 * Save_DEM
 *
 * Description: 
 *
 * Reports:      OK
 */

/* Save_DEMMain  -  codel EXEC of Save_DEM
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
Save_DEMMain(SAVE_DEM_I *Save_DEM_I, int *report)
{
  // check if the DEM log is not empty
  if(logDEM.size() == 0)
    {
      cerr << "Cannot save a DEM. The DEM log is empty." << endl;
      return ETHER;
    }

  // select a DEM in the log
  int num = (int)(logDEM.size()*0.5);
  // get back the corresponding DEM
  std::list<DEM*>::iterator it = logDEM.begin();
  for(int i=0; i<num; i++) it++;
  DEM* DEM_num = *it;

  // check if the selected DEM is not null
  if(DEM_num == NULL)
    {
      cerr << "Cannot save a DEM. The DEM is null." << endl;
      return ETHER;
    }

  // save DEM computed in the middle of the DEM log, and save it into an output file
  int res_save = DEM_num->save(Save_DEM_I->filename);
  if(res_save != 0)
    {
      cerr << "Cannot save a DEM. The DEM is null." << endl;
      return ETHER;
    }

  return ETHER;
}
